/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.tools.edit.behaviour;

import java.util.HashMap;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import net.refractions.udig.core.IProvider;
import net.refractions.udig.project.command.UndoableComposite;
import net.refractions.udig.project.command.UndoableMapCommand;
import net.refractions.udig.project.ui.AnimationUpdater;
import net.refractions.udig.project.ui.render.displayAdapter.MapMouseEvent;
import net.refractions.udig.tool.edit.internal.Messages;
import net.refractions.udig.tools.edit.EditPlugin;
import net.refractions.udig.tools.edit.EditState;
import net.refractions.udig.tools.edit.EditToolHandler;
import net.refractions.udig.tools.edit.EventBehaviour;
import net.refractions.udig.tools.edit.EventType;
import net.refractions.udig.tools.edit.LockingBehaviour;
import net.refractions.udig.tools.edit.animation.MessageBubble;
import net.refractions.udig.tools.edit.commands.DrawSnapAreaCommand;
import net.refractions.udig.tools.edit.commands.SetEditStateCommand;
import net.refractions.udig.tools.edit.commands.SnapToVertexCommand;
import net.refractions.udig.tools.edit.preferences.PreferenceUtil;
import net.refractions.udig.tools.edit.support.EditBlackboard;
import net.refractions.udig.tools.edit.support.EditGeom;
import net.refractions.udig.tools.edit.support.Point;
import net.refractions.udig.tools.edit.support.Selection;
import net.refractions.udig.tools.edit.support.SnapBehaviour;
import net.refractions.udig.tools.edit.validator.LegalShapeValidator;

/**
 * Mode that moves Vertices
 * <p>
 * Requirements:
 * <ul>
 * <li>currentGeom!=null;</li>
 * <li>currentState is MODIFIED or NONE </li>
 * <li>eventType is DRAGGED</li>
 * <li>no modifiers</li>
 * <li>{@link net.refractions.udig.tools.edit.MouseTracker#getDragStarted()} is over selected
 * vertex (if not all vertices are selected)</li>
 * <li>If all vertices are selected then
 * {@link net.refractions.udig.tools.edit.MouseTracker#getDragStarted()} must be within the
 * geometry.
 * </ul>
 * </p>
 * <p>
 * Action:
 * <ul>
 * <li>move the vertices selected by {@link VertexSelectorBehaviour} to location in
 * {@link net.refractions.udig.project.ui.render.displayAdapter.MapMouseEvent}</li>
 * <li>Locks the EditTool handler until mouse is released so other behaviours won't interfere</li>
 * </ul>
 * </p>
 * 
 * @author Jesse
 * @since 1.1.0
 */
public class MoveVertexBehaviour implements EventBehaviour, LockingBehaviour {

    PositionTracker tracker;

    public boolean isValid( EditToolHandler handler, MapMouseEvent e, EventType eventType ) {
        boolean isLegalState = handler.getCurrentState() == EditState.MODIFYING
                || handler.getCurrentState() == EditState.NONE
                || handler.getCurrentState() == EditState.MOVING;
        boolean isEventDragged = eventType == EventType.DRAGGED;
        boolean noModifiersDown = !e.modifiersDown();
        boolean button1IsDown = (e.buttons ^ MapMouseEvent.BUTTON1) == 0;
        boolean currentGeomNotNull = handler.getCurrentGeom() != null;

        return isLegalState && isEventDragged && noModifiersDown && button1IsDown
                && currentGeomNotNull
                && (handler.isLockOwner(this) || startedOverSelectedVertex(handler));
    }

    private boolean startedOverSelectedVertex( EditToolHandler handler ) {
        // if tracker!=null then the move has started so the selection is no longer
        // at the same spot as the selected point so return true
        if (tracker != null)
            return true;
        Point started = handler.getMouseTracker().getDragStarted();
        EditBlackboard editBlackboard = handler.getEditBlackboard(handler.getEditLayer());
        Point point = handler.getEditBlackboard(handler.getEditLayer())
                .overVertex(Point.valueOf(started.getX(), started.getY()),
                        PreferenceUtil.instance().getVertexRadius());
        if (point == null) {
            point = Point.valueOf(started.getX(), started.getY());
        }
        return getPointsToMove(handler, editBlackboard).contains(point);
    }

    DrawSnapAreaCommand drawSnapArea;

    public UndoableMapCommand getCommand( EditToolHandler handler, MapMouseEvent e,
            EventType eventType ) {
        if (handler.getCurrentState() != EditState.MOVING)
            handler.setCurrentState(EditState.MOVING);

        if (!isValid(handler, e, eventType))
            throw new IllegalArgumentException("Not valid state", new Exception()); //$NON-NLS-1$

        EditBlackboard editBlackboard2 = handler.getEditBlackboard(handler.getEditLayer());
        editBlackboard2.startBatchingEvents();
        try {
            if (tracker == null) {
                handler.lock(this);
                Point closestPoint = editBlackboard2.overVertex(Point.valueOf(e.x, e.y),
                        PreferenceUtil.instance().getVertexRadius(), false);
                Map<EditGeom, Boolean> changedStatus=new HashMap<EditGeom, Boolean>();
                for( EditGeom geom : editBlackboard2.getGeoms() ) {
                    changedStatus.put(geom, geom.isChanged());
                }
                tracker = new PositionTracker(closestPoint, handler.getMouseTracker()
                        .getDragStarted(), getPointsToMove(handler, editBlackboard2), 
                        changedStatus);
                handler.getBehaviours().add(tracker);
                if (isSnappingValid() && PreferenceUtil.instance().getSnapBehaviour()!=SnapBehaviour.GRID ) {
                        drawSnapArea = new DrawSnapAreaCommand(tracker);
                    handler.getContext().getViewportPane().addDrawCommand(drawSnapArea);

                }
            }

            if (tracker.lastPoint == null) {
                tracker.lastPoint = handler.getMouseTracker().getDragStarted();
            }

            Point point = Point.valueOf(e.x, e.y);

            int deltaX = point.getX() - tracker.lastPoint.getX(), deltaY = point.getY()
                    - tracker.lastPoint.getY();

            doMove(deltaX, deltaY, handler, editBlackboard2, tracker.selection);

            tracker.lastPoint = point;
            return null;
        } finally {
            editBlackboard2.fireBatchedEvents();
            handler.repaint();
        }
    }

    protected void doMove( int deltaX, int deltaY, EditToolHandler handler,
            EditBlackboard editBlackboard2, Selection selectionToMove ) {
        editBlackboard2.moveSelection(deltaX, deltaY, selectionToMove);
    }

    /**
     * Returns true if snapping should be used.
     * 
     * @return
     */
    protected boolean isSnappingValid() {
        return PreferenceUtil.instance().getSnapBehaviour()!=SnapBehaviour.OFF;
    }

    /**
     * Returns the points that will be moved.
     * 
     * @param handler
     * @return the points that will be moved.
     */
    protected Selection getPointsToMove( EditToolHandler handler, EditBlackboard blackboard ) {
        return blackboard.getSelection();
    }

    public void handleError( EditToolHandler handler, Throwable error, UndoableMapCommand command ) {
        EditPlugin.log("", error); //$NON-NLS-1$
    }

    public class PositionTracker implements LockingBehaviour, IProvider<Point> {
        Point lastPoint;
        private Point start;
        private Selection selection;
        private Map<EditGeom, Boolean> dirtyStatesBeforeMove;

        public PositionTracker( Point closestPoint, Point dragStarted, Selection selection,
                Map<EditGeom, Boolean> dirtyStatesBeforeMove ) {
            this.selection = selection;
            lastPoint = closestPoint != null ? closestPoint : dragStarted;
            this.start = lastPoint;
            this.dirtyStatesBeforeMove=dirtyStatesBeforeMove;
            
        }

        public boolean isValid( EditToolHandler handler, MapMouseEvent e, EventType eventType ) {
            return eventType == EventType.RELEASED;
        }

        public UndoableMapCommand getCommand( EditToolHandler handler, MapMouseEvent e,
                EventType eventType ) {

            if (drawSnapArea != null) {
                drawSnapArea.setValid(false);
                drawSnapArea = null;
            }
            handler.getBehaviours().remove(this);
            tracker = null;
            handler.unlock(this);
            
            LegalShapeValidator validator = new LegalShapeValidator();
            String errorMessage = validator.isValid(handler, e, eventType);
            if( errorMessage==null ){
                UndoableComposite command = new UndoableComposite();
                command.getCommands().add(
                        new SnapToVertexCommand(lastPoint, this.selection, handler, start,
                                EditState.MODIFYING, isSnappingValid()));
                command.getFinalizerCommands().add(
                        new SetEditStateCommand(handler, EditState.MODIFYING));
                return command;
            }else{
                MessageBubble bubble=new MessageBubble(e.getPoint().x, e.getPoint().y,
                        Messages.MoveVertexBehaviour_validationError+"\n"+errorMessage, //$NON-NLS-1$ 
                        PreferenceUtil.instance().getMessageDisplayDelay());
                AnimationUpdater.runTimer(handler.getContext().getMapDisplay(), bubble);
                EditBlackboard editBlackboard = handler.getEditBlackboard(handler.getEditLayer());
                doMove(start.getX()-lastPoint.getX(), start.getY()-lastPoint.getY(), 
                        handler, editBlackboard, selection);
                Set<Entry<EditGeom, Boolean>> entries = dirtyStatesBeforeMove.entrySet();
                for( Entry<EditGeom, Boolean> entry : entries ) {
                    entry.getKey().setChanged(entry.getValue());
                }
                return null;
            }

        }

        public void handleError( EditToolHandler handler, Throwable error,
                UndoableMapCommand command ) {
            EditPlugin.log("", error); //$NON-NLS-1$
        }

        public Point get() {
            return lastPoint;
        }

        public Object getKey( EditToolHandler handler ) {
            return MoveVertexBehaviour.this;
        }

    }

    public Object getKey( EditToolHandler handler ) {
        return this;
    }

}
