/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.tools.edit.behaviour;

import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.command.UndoableComposite;
import net.refractions.udig.project.command.UndoableMapCommand;
import net.refractions.udig.project.ui.render.displayAdapter.MapMouseEvent;
import net.refractions.udig.tools.edit.EditPlugin;
import net.refractions.udig.tools.edit.EditState;
import net.refractions.udig.tools.edit.EditToolHandler;
import net.refractions.udig.tools.edit.EventBehaviour;
import net.refractions.udig.tools.edit.EventType;
import net.refractions.udig.tools.edit.commands.AddMissingGeomsCommand;
import net.refractions.udig.tools.edit.commands.RemoveEditGeomCommand;
import net.refractions.udig.tools.edit.commands.SelectGeometryCommand;
import net.refractions.udig.tools.edit.commands.SetCurrentGeomCommand;
import net.refractions.udig.tools.edit.commands.SetEditFeatureCommand;
import net.refractions.udig.tools.edit.preferences.PreferenceUtil;
import net.refractions.udig.tools.edit.support.ClosestEdge;
import net.refractions.udig.tools.edit.support.EditBlackboard;
import net.refractions.udig.tools.edit.support.EditGeom;
import net.refractions.udig.tools.edit.support.EditUtils;
import net.refractions.udig.tools.edit.support.Point;
import net.refractions.udig.tools.edit.support.PrimitiveShape;
import net.refractions.udig.tools.edit.support.PrimitiveShapeIterator;
import net.refractions.udig.tools.edit.support.ShapeType;

import org.geotools.filter.FilterType;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Polygon;

/**
 * <p>
 * Requirements:
 * <ul>
 * <li>state==MODIFYING or NONE</li>
 * <li>event type == RELEASED</li>
 * <li>only a single modifier may be down</li>
 * <li>shift and ctrl are only legal modifiers</li>
 * <li>button1 must be the button that was released</li>
 * <li>no buttons may be down</li>
 * <li>Not over currently selected geometry</li>
 * </ul>
 * </p>
 * <p>
 * Action:
 * <ul>
 * <li>If no currently selected geometry && mouse is over feature then:
 * <ul>
 * <li>If the feature's geometry isn't of the type specified in constructor return or:</li>
 * <li>Clear Editblackboard and set feature's geometry on blackboard.</li>
 * <li>set the current geom to be the geometry selected</li>
 * <li>set the current Edit feature to be selected feature</li>
 * <li>set mode to be Modified?</li>
 * </ul>
 * </li>
 * <li>If there is a currently selected geometry && mouse is over a different feature:
 * <ul>
 * <li>If no modifiers select feature as described above</li>
 * <li>if control is down add the feature to the Editblackboard or remove it if it is already on
 * the EditBlackboard</li>
 * <li>if shift is down add the feature to the Editblackboard or do nothing if it is already on the
 * EditBlackboard</li>
 * </ul>
 * <li>If there is a currently selected geometry && mouse is not over a different feature:
 * <ul>
 * <li>Run acceptBehaviours</li>
 * <li>deselect EditFeature</li>
 * <li>deselect current EditGeom</li>
 * <li>clear Editblackboard</li>
 * <li>set current state to NONE</li>
 * </ul>
 * </li>
 * </ul>
 * </p>
 * 
 * @author jones
 * @since 1.1.0
 */
public class SelectGeometryBehaviour implements EventBehaviour {

    private Class[] acceptableClasses;
    private boolean treatUnknownGeomsAsPolygon;
    private short filterType;
    private boolean permitClear;
    private boolean onlyAdd;
    private ShapeType toCreate;

    /**
     * Create instance
     * 
     * @param acceptableClasses used to determine if a feature can be selected. If point is not in
     *        array then point geometries can not be selected.
     * @param filterType one of the constants in {@link FilterType} that start with GEOMETRY_
     */
    public SelectGeometryBehaviour( Class[] acceptableClasses, short filterType ) {
        Class[] c=new Class[0];
        if( acceptableClasses!=null ){
            c=new Class[acceptableClasses.length];
            System.arraycopy(acceptableClasses, 0, c, 0, c.length);
        }

        this.acceptableClasses = c;
        this.treatUnknownGeomsAsPolygon = false;
        for( Class<Geometry> class1 : c ) {
            if (class1.isAssignableFrom(Polygon.class)
                    || class1.isAssignableFrom(MultiPolygon.class)) {
                treatUnknownGeomsAsPolygon = true;
                break;
            }
        }
        this.filterType = filterType;
        onlyAdd=false;
        permitClear=true;
    }
    
    public void setCreateGeomOnNoneSelect(ShapeType toCreate2){
        this.toCreate=toCreate2;
    }

    public boolean isValid( EditToolHandler handler, MapMouseEvent e, EventType eventType ) {
        boolean legalState = handler.getCurrentState() == EditState.NONE
                || handler.getCurrentState() == EditState.MODIFYING;
        boolean releaseButtonState = eventType == EventType.RELEASED;
        boolean twoModifiersDown = e.isShiftDown() && e.isModifierDown(MapMouseEvent.MOD1_DOWN_MASK);
        boolean singleModifierDown = !(twoModifiersDown);
        boolean altUp = !e.isAltDown();
        boolean legalButton = e.button == MapMouseEvent.BUTTON1;
        boolean noPressedButtons = e.buttons == MapMouseEvent.NONE;
        if (!(legalState && releaseButtonState && altUp && singleModifierDown && legalButton && noPressedButtons))
            return false;

        if (e.isModifierDown(MapMouseEvent.MOD1_DOWN_MASK))
            return true;

        if (handler.getCurrentGeom() == null)
            return true;
        
        if (!handler.getCurrentShape().contains(Point.valueOf(e.x, e.y), treatUnknownGeomsAsPolygon))
            return true;

        return countOnBlackboard(handler, e) > 1; //click was within the current blackboard selection
    }

    public UndoableMapCommand getCommand( final EditToolHandler handler, final MapMouseEvent e,
            EventType eventType ) {
        if (!isValid(handler, e, eventType))
            throw new IllegalArgumentException("Behaviour is not valid for the current state"); //$NON-NLS-1$
            EditBlackboard editBlackboard = handler.getEditBlackboard(handler.getEditLayer());
            List<EditGeom> intersectingGeoms = EditUtils.instance.getIntersectingGeom(
                    editBlackboard, Point.valueOf(e.x, e.y), treatUnknownGeomsAsPolygon);
            if (e.isModifierDown(MapMouseEvent.MOD1_DOWN_MASK) && !intersectingGeoms.isEmpty()) {
                return new RemoveEditGeomCommand(handler, intersectingGeoms);
            } else if (e.isShiftDown() && !intersectingGeoms.isEmpty()) {
                return null;
            }

            PrimitiveShape newShape=findOnBlackboard(handler, e);

            if( newShape!=null && newShape!=handler.getCurrentShape() ){
                List<UndoableMapCommand> commands = new ArrayList<UndoableMapCommand>();
                commands.add(new AddMissingGeomsCommand(handler, e, acceptableClasses, filterType, onlyAdd));
                commands.add(new SetCurrentGeomCommand(handler, newShape));
                commands.add(new SetEditFeatureCommand(handler, e, newShape));
                UndoableComposite undoableComposite = new UndoableComposite(
                        commands);
                return undoableComposite;
            }
            
            SelectGeometryCommand selectGeometryCommand = new SelectGeometryCommand(handler,e, acceptableClasses, 
                    filterType, permitClear, onlyAdd);
            selectGeometryCommand.setCreateOnNoSelect(toCreate);
            return selectGeometryCommand;
    }

    private PrimitiveShape findOnBlackboard( EditToolHandler handler, MapMouseEvent e ) {
        //for overlapping geometries, select a different one on each click
        boolean cycleGeom = false;
        //set when the currently selected feature has been found
        boolean selectedFound = false;
        //for returning to the first match when we have the last match selected and click once more
        PrimitiveShape firstMatch = null; 
        if (handler.getCurrentShape() != null && handler.getCurrentShape().contains(e.x, e.y) && countOnBlackboard(handler, e) > 1) {
            cycleGeom = true;
        }
            
        ILayer editLayer = handler.getEditLayer();
        List<EditGeom> geoms = handler.getEditBlackboard(editLayer).getGeoms();
        for( EditGeom geom : geoms ) {
            PrimitiveShapeIterator iter=PrimitiveShapeIterator.getPathIterator(geom.getShell());
            if( iter.toShape().contains(e.x,e.y) ){
                if (cycleGeom) {
                    if (selectedFound) { //first match after the currently selected one
                        return geom.getShell();
                    }
                    if (geom == handler.getCurrentGeom()) { //currently selected geom
                        selectedFound = true;
                    }
                    if (firstMatch == null) { //first matching geom
                        firstMatch = geom.getShell();
                    }
                } else {
                    return geom.getShell();
                }
            }
            if (!cycleGeom) {
                Class type = editLayer.getSchema().getDefaultGeometry().getType();
                boolean polygonLayer=Polygon.class.isAssignableFrom(type) || MultiPolygon.class.isAssignableFrom(type);
                ClosestEdge edge = geom.getShell().getClosestEdge(Point.valueOf(e.x, e.y), polygonLayer);
                if (edge != null && 
                        edge.getDistanceToEdge() <= PreferenceUtil.instance().getVertexRadius()){
                    return geom.getShell();
                }
            }
        }
        if (cycleGeom) //selected geom was the last one, therefore select the first match
            return firstMatch;
        else
            return null;
    }

    private int countOnBlackboard( EditToolHandler handler, MapMouseEvent e ) {
        List<EditGeom> geoms = handler.getEditBlackboard(handler.getEditLayer()).getGeoms();
        int count = 0;
        for( EditGeom geom : geoms ) {
            PrimitiveShapeIterator iter=PrimitiveShapeIterator.getPathIterator(geom.getShell());
            if( iter.toShape().contains(e.x,e.y) ){
                count++;
            }
        }
        return count;
    }
    
    public void handleError( EditToolHandler handler, Throwable error, UndoableMapCommand command ) {
        EditPlugin.log("", error); //$NON-NLS-1$
    }

    /**
     * @return Returns the acceptableClasses.
     */
    public Class[] getAcceptableClasses() {
        Class[] c=new Class[acceptableClasses.length];
        System.arraycopy(acceptableClasses, 0, c, 0, c.length);
        return c;
    }

    /**
     * @param acceptableClasses The acceptableClasses to set.
     */
    public void setAcceptableClasses( Class[] acceptableClasses ) {
        Class[] c=new Class[0];
        if( acceptableClasses!=null ){
            c=new Class[acceptableClasses.length];
            System.arraycopy(acceptableClasses, 0, c, 0, c.length);
        }
        this.acceptableClasses = c;
    }

    /**
     * @return Returns the filterType.
     */
    public short getFilterType() {
        return this.filterType;
    }

    /**
     * @param filterType The filterType to set.
     */
    public void setFilterType( short filterType ) {
        this.filterType = filterType;
    }

    /**
     * @return Returns the onlyAdd.
     */
    public boolean isOnlyAdd() {
        return this.onlyAdd;
    }

    /**
     * @param onlyAdd The onlyAdd to set.
     */
    public void setOnlyAdd( boolean onlyAdd ) {
        this.onlyAdd = onlyAdd;
    }

    /**
     * @return Returns the permitClear.
     */
    public boolean isPermitClear() {
        return this.permitClear;
    }

    /**
     * @param permitClear The permitClear to set.
     */
    public void setPermitClear( boolean permitClear ) {
        this.permitClear = permitClear;
    }

    /**
     * @return Returns the treatUnkownGeomsAsPolygon.
     */
    public boolean isTreatUnknownGeomsAsPolygon() {
        return this.treatUnknownGeomsAsPolygon;
    }

    /**
     * @param treatUnknownGeomsAsPolygon The treatUnknownGeomsAsPolygon to set.
     */
    public void setTreatUnknownGeomsAsPolygon( boolean treatUnknownGeomsAsPolygon ) {
        this.treatUnknownGeomsAsPolygon = treatUnknownGeomsAsPolygon;
    }

}
