/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.tools.edit.behaviour;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import net.refractions.udig.core.IBlockingProvider;
import net.refractions.udig.core.StaticBlockingProvider;
import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.command.UndoableComposite;
import net.refractions.udig.project.command.UndoableMapCommand;
import net.refractions.udig.project.command.factory.EditCommandFactory;
import net.refractions.udig.project.command.provider.FIDFeatureProvider;
import net.refractions.udig.project.internal.commands.edit.DeleteFeatureCommand;
import net.refractions.udig.project.render.displayAdapter.IMapDisplay;
import net.refractions.udig.project.ui.IAnimation;
import net.refractions.udig.project.ui.commands.DrawCommandFactory;
import net.refractions.udig.tool.edit.internal.Messages;
import net.refractions.udig.tools.edit.Behaviour;
import net.refractions.udig.tools.edit.EditPlugin;
import net.refractions.udig.tools.edit.EditState;
import net.refractions.udig.tools.edit.EditToolHandler;
import net.refractions.udig.tools.edit.animation.GeometryOperationAnimation;
import net.refractions.udig.tools.edit.commands.AddVertexCommand;
import net.refractions.udig.tools.edit.commands.CreateAndSetNewFeature;
import net.refractions.udig.tools.edit.commands.CreateOrSetFeatureCommand;
import net.refractions.udig.tools.edit.commands.SetEditGeomChangedStateCommand;
import net.refractions.udig.tools.edit.commands.SetEditStateCommand;
import net.refractions.udig.tools.edit.support.EditGeom;
import net.refractions.udig.tools.edit.support.EditGeomPathIterator;
import net.refractions.udig.tools.edit.support.GeometryCreationUtil;
import net.refractions.udig.tools.edit.support.IsBusyStateProvider;
import net.refractions.udig.tools.edit.support.PrimitiveShape;
import net.refractions.udig.tools.edit.support.ShapeType;

import org.geotools.feature.Feature;
import org.geotools.feature.IllegalAttributeException;

import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.LineString;
import com.vividsolutions.jts.geom.LinearRing;
import com.vividsolutions.jts.geom.MultiLineString;
import com.vividsolutions.jts.geom.MultiPoint;
import com.vividsolutions.jts.geom.MultiPolygon;
import com.vividsolutions.jts.geom.Point;
import com.vividsolutions.jts.geom.Polygon;

/**
 * Creates a geometry from the currentGeom in the {@link EditToolHandler}.
 * <p>
 * Requirements:
 * <ul>
 * <li>handler.getCurrentGeom() != null</li>
 * <li>EditManager has a edit feature and a edit layer</li>
 * <li>Geometry can be assigned to feature's geometry</li>
 * </ul>
 * </p>
 * <p>
 * Action:
 * <ul>
 * <li>Creates a geometry of the indicated type.</li>
 * <li>Set the edit feature's default geometry to be the created geometry if it is not null</li>
 * <li>Also writes the changes to other geoms in the blackboard. The type created for those depends
 * on the what the ShapeType of those geometries are.</li>
 * </ul>
 * </p>
 * 
 * @author jones
 * @since 1.1.0
 */
public class WriteChangesBehaviour implements Behaviour {

    private final Class< ? extends Geometry> geomToCreate;
    private boolean updateBlackboard=true;

    /**
     * @param geomToCreate must be polygon, point, multipoint, linestring or linearring
     */
    public WriteChangesBehaviour( Class< ? extends Geometry> geomToCreate ) {
        if (geomToCreate != Polygon.class && geomToCreate != Point.class
                && geomToCreate != LineString.class && geomToCreate != LinearRing.class)
            throw new IllegalArgumentException(
                    "Must be one of Polygon, Point, LineString, MultiPoint or LinearRing"); //$NON-NLS-1$

        this.geomToCreate = geomToCreate;
    }

    public boolean isValid( EditToolHandler handler ) {
        EditGeom currentGeom = handler.getCurrentGeom();
        
        boolean currentGeomNotNull = currentGeom != null;
        return currentGeomNotNull
                && typeCanBeAssignedToLayer(handler.getEditLayer().getSchema()
                        .getDefaultGeometry().getType());
    }

    /**
     * @param type
     * @return
     */
    @SuppressWarnings("unchecked")
    private boolean typeCanBeAssignedToLayer( Class type ) {
        if (type.isAssignableFrom(geomToCreate))
            return true;
        if (geomToCreate == Polygon.class && type.isAssignableFrom(MultiPolygon.class)) {
            return true;
        } else if (geomToCreate == LinearRing.class && type.isAssignableFrom(MultiLineString.class)) {
            return true;
        } else if (geomToCreate == LineString.class && type.isAssignableFrom(MultiLineString.class)) {
            return true;
        } else if (geomToCreate == Point.class && type.isAssignableFrom(MultiPoint.class)) {
            return true;
        }
        return false;
    }

    @SuppressWarnings({"deprecation","unchecked"}) 
    public UndoableMapCommand getCommand( EditToolHandler handler ) {
        if (!isValid(handler))
            throw new IllegalArgumentException("Not in a valid state for this to run"); //$NON-NLS-1$
        Map<String, GeometryCreationUtil.Bag> idToGeom = GeometryCreationUtil.createAllGeoms(
                handler.getCurrentGeom(), geomToCreate, handler.getEditLayer()
                        .getSchema().getDefaultGeometry());

        EditState oldState = handler.getCurrentState();
        
        IMapDisplay display = handler.getContext().getMapDisplay();
        DrawCommandFactory drawfactory = handler.getContext().getDrawFactory();
        List<UndoableMapCommand> commands = new ArrayList<UndoableMapCommand>();
        commands.add(new SetEditStateCommand(handler, EditState.COMMITTING));
        
        ILayer layer = handler.getEditLayer();
        EditCommandFactory editFactory = handler.getContext().getEditFactory();
        Set<Entry<String, GeometryCreationUtil.Bag>> entries = idToGeom.entrySet();
        for( Entry<String, GeometryCreationUtil.Bag> entry : entries ) {
            EditGeom editGeom = entry.getValue().geom;
            List<Geometry> geoms = entry.getValue().jts;

            UndoableMapCommand command;
            Geometry geom = GeometryCreationUtil.ceateGeometryCollection(geoms, layer.getSchema()
                    .getDefaultGeometry().getType());

            if( geom==null ){
                IBlockingProvider<ILayer> layerProvider = new StaticBlockingProvider<ILayer>(layer);
                return new DeleteFeatureCommand(new FIDFeatureProvider(entry.getKey(), layerProvider), 
                        layerProvider);
            }
            
            if( updateBlackboard )
                updateBlackboardGeometry(handler, editGeom, geom, commands);

            GeometryOperationAnimation animation = new GeometryOperationAnimation(
                    EditGeomPathIterator.getPathIterator(editGeom).toShape(),
                    new IsBusyStateProvider(handler));

            commands.add(drawfactory.createStartAnimationCommand(display, Collections
                    .singletonList((IAnimation) animation)));

            if (editGeom == handler.getCurrentGeom()) {
                command = editFactory.createSetGeometryCommand(geom);
            } else {
                UndoableComposite composite = new UndoableComposite();
                composite.getCommands()
                        .add(
                                new CreateOrSetFeatureCommand(editGeom.getFeatureIDRef().get(),
                                        layer, geom));
                command = composite;
            }

            if (editGeom == handler.getCurrentGeom()
                    && handler.getContext().getEditManager().getEditFeature() == null) {
                int i = layer.getSchema().getAttributeCount();

                Feature feature;
                try {
                    feature = layer.getSchema().create(new Object[i]);
                    feature.setDefaultGeometry(geom);
                } catch (IllegalAttributeException e) {
                    throw (RuntimeException) new RuntimeException().initCause(e);
                }
                commands.add(new CreateAndSetNewFeature(handler.getCurrentGeom(), feature, layer));
            }else{
                // not creating it so don't need to set it.
                commands.add(command);
            }
            commands.add(new SetEditGeomChangedStateCommand(editGeom, false));
            commands.add(drawfactory.createStopAnimationCommand(display, Collections
                    .singletonList((IAnimation) animation)));
        }

        UndoableComposite composite = new UndoableComposite(commands);
        composite.setName(Messages.WriteChangesBehaviour_commandName);
        composite.getFinalizerCommands().add(new SetEditStateCommand(handler, EditState.MODIFYING));

        handler.setCurrentState(oldState);
        
        return composite;
    }

    private void updateBlackboardGeometry( EditToolHandler handler, EditGeom editGeom, Geometry geom, List<UndoableMapCommand> commands ) {
        if( handler.getCurrentGeom()==editGeom){
            if( Polygon.class.isAssignableFrom(geomToCreate) ){
                for( PrimitiveShape shape : editGeom ) {
                    if( shape.getNumPoints()>0 && !shape.getPoint(0).equals(shape.getPoint(shape.getNumPoints()-1)))
                        commands.add(new AddVertexCommand(handler, editGeom.getEditBlackboard(), shape.getPoint(0)) );
                }
            }
        }else{
            if( editGeom.getShapeType()==ShapeType.POLYGON || (editGeom.getShapeType()==ShapeType.UNKNOWN && Polygon.class.isAssignableFrom(geomToCreate)) ){
                for( PrimitiveShape shape : editGeom ) {
                    if( shape.getNumPoints()>0 && !shape.getPoint(0).equals(shape.getPoint(shape.getNumPoints()-1)))
                        commands.add(new AddVertexCommand(handler, editGeom.getEditBlackboard(), shape.getPoint(0)) );
                }
            }            
        }
    }
    
    public void handleError( EditToolHandler handler, Throwable error, UndoableMapCommand command ) {
        EditPlugin.log("", error); //$NON-NLS-1$
    }

    /**
     * @param updateBlackboard The updateBlackboard to set.
     */
    public void setUpdateBlackboard( boolean updateBlackboard ) {
        this.updateBlackboard = updateBlackboard;
    }
}
