package net.refractions.udig.tools.edit.commands;

import java.util.ArrayList;
import java.util.List;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.command.AbstractCommand;
import net.refractions.udig.project.command.UndoableComposite;
import net.refractions.udig.project.command.UndoableMapCommand;
import net.refractions.udig.project.internal.ProjectPlugin;
import net.refractions.udig.project.ui.render.displayAdapter.MapMouseEvent;
import net.refractions.udig.project.ui.tool.IToolContext;
import net.refractions.udig.tool.edit.internal.Messages;
import net.refractions.udig.tools.edit.EditPlugin;
import net.refractions.udig.tools.edit.EditToolHandler;
import net.refractions.udig.tools.edit.support.EditBlackboard;

import org.eclipse.core.runtime.IProgressMonitor;
import org.geotools.data.FeatureSource;
import org.geotools.data.FeatureStore;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.filter.BBoxExpression;
import org.geotools.filter.Filter;
import org.geotools.filter.FilterFactory;
import org.geotools.filter.FilterFactoryFinder;
import org.geotools.filter.FilterType;
import org.geotools.filter.GeometryFilter;
import org.geotools.geometry.jts.JTS;
import org.geotools.geometry.jts.ReferencedEnvelope;
import org.opengis.referencing.operation.MathTransform;

import com.vividsolutions.jts.geom.Envelope;
import com.vividsolutions.jts.geom.Geometry;

/**
 * Adds any geometries containing the given point which are not already on the edit blackboard.
 * 
 * @author chorner
 * @since 1.1.0
 */
public class AddMissingGeomsCommand extends AbstractCommand implements UndoableMapCommand {

    private EditToolHandler handler;
    private Class[] acceptableClasses;
    private boolean onlyAdd;
    private MapMouseEvent event;
    private short filterType;

    private UndoableMapCommand command;

    public AddMissingGeomsCommand( EditToolHandler handler, MapMouseEvent e,
            Class[] acceptableClasses, short filterType, boolean onlyAdd ) {
        this.handler = handler;
        this.event = e;
        Class[] c=new Class[0];
        if( acceptableClasses!=null ){
            c=new Class[acceptableClasses.length];
            System.arraycopy(acceptableClasses, 0, c, 0, c.length);
        }
        this.acceptableClasses = c;

        this.filterType = filterType;
        this.onlyAdd = onlyAdd;
    }

    public void run( IProgressMonitor monitor ) throws Exception {
        if (command == null) {
            EditBlackboard editBlackboard = handler.getEditBlackboard(handler.getEditLayer());
            editBlackboard.startBatchingEvents();
            try {
                Feature feature = null;
                ILayer editLayer = handler.getEditLayer();
                try {
                    FeatureStore store = editLayer.getResource(FeatureStore.class, null);
                    ReferencedEnvelope bbox = handler.getContext().getBoundingBox(event.getPoint(),
                            7);
                    Filter createBBoxFilter = createBBoxFilter(bbox, editLayer, filterType);
                    FeatureCollection collection = store.getFeatures(createBBoxFilter);
                    FeatureIterator reader = collection.features();
                    try {
                        boolean hasFeature = false;
                        try {
                            hasFeature = reader.hasNext();
                        } catch (Exception e) {
                            reader.close();
                            createBBoxFilter = createBBoxFilter(bbox, editLayer, FilterType.GEOMETRY_BBOX);
                            collection = store.getFeatures(createBBoxFilter);
                            reader = collection.features();
                            hasFeature = reader.hasNext();
                        }
                        if (hasFeature) {
                            List<UndoableMapCommand> commands = new ArrayList<UndoableMapCommand>();
                            while( reader.hasNext() ) {
                                feature = reader.next();
                                for( Class<Geometry> clazz : acceptableClasses ) {
                                    if (onlyAdd || handler.getCurrentGeom() != null
                                            && (event.isShiftDown() || event.isControlDown())) {
                                        setAndRun(monitor, new AddGeomCommand(handler, feature, null));
                                    } else {
                                        if (clazz.isAssignableFrom(feature.getDefaultGeometry()
                                                .getClass())) {
                                            if (!editBlackboard.contains(feature.getID())) {
                                                commands.add(new AddGeomCommand(handler, feature, null));
                                                EditPlugin
                                                    .trace(
                                                            EditPlugin.SELECTION,
                                                            "Feature is one of the acceptable classes " + feature.getID(), null); //$NON-NLS-1$
                                            }
                                        } else {
                                            EditPlugin
                                                    .trace(
                                                            EditPlugin.SELECTION,
                                                            "Feature is not one of the acceptable classes " + feature.getID(), null); //$NON-NLS-1$
                                        }
                                    }
                                }
                            }
                            UndoableComposite undoableComposite = new UndoableComposite(
                                    commands);
                            undoableComposite.setName(Messages.AddMissingGeomsCommand_name);
                            setAndRun(monitor, undoableComposite);
                        }
                    } finally {
                        collection.close(reader);
                    }
                } catch (Exception e2) {
                    EditPlugin.log("", e2); //$NON-NLS-1$
                }
            } finally {
                editBlackboard.fireBatchedEvents();
            }
        } else {
            command.run(monitor);
        }
    }

    public String getName() {
        return Messages.AddMissingGeomsCommand_name;
    }

    public void rollback( IProgressMonitor monitor ) throws Exception {
        command.rollback(monitor);
    }

    /**
     * Creates A geometry filter for the given layer.
     * 
     * @param boundingBox in the same crs as the viewport model.
     * @return a Geometry filter in the correct CRS or null if an exception occurs.
     */
    public Filter createBBoxFilter( Envelope boundingBox, ILayer layer, short filterType ) {
        FilterFactory factory = FilterFactoryFinder.createFilterFactory();
        GeometryFilter bboxFilter = null;
        if (!layer.hasResource(FeatureSource.class))
            return Filter.ALL;
        try {

            Envelope bbox;
            try {
                MathTransform transform = layer.mapToLayerTransform();
                bbox = JTS.transform(boundingBox, transform);
            } catch (Exception e) {
                bbox = boundingBox;
            }
            BBoxExpression bb = factory.createBBoxExpression(bbox);
            bboxFilter = factory.createGeometryFilter(filterType);
            bboxFilter.addRightGeometry(bb);

            String geom = layer.getSchema().getDefaultGeometry().getName();

            bboxFilter.addLeftGeometry(factory.createAttributeExpression(geom));
        } catch (Exception e) {
            ProjectPlugin.getPlugin().log(e);
        }
        return bboxFilter;
    }
    
    private void setAndRun (IProgressMonitor monitor, UndoableMapCommand undoableComposite) throws Exception {
        command = undoableComposite;
        command.setMap(getMap());
        command.run(monitor);
    }
}
