/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.tools.edit.commands;

import java.awt.Color;
import java.awt.Rectangle;
import java.util.Iterator;

import net.refractions.udig.project.ui.commands.AbstractDrawCommand;
import net.refractions.udig.project.ui.render.displayAdapter.MapMouseEvent;
import net.refractions.udig.project.ui.render.displayAdapter.MapMouseMotionListener;
import net.refractions.udig.project.ui.render.displayAdapter.ViewportPane;
import net.refractions.udig.tools.edit.EditPlugin;
import net.refractions.udig.tools.edit.EditToolHandler;
import net.refractions.udig.tools.edit.preferences.PreferenceConstants;
import net.refractions.udig.tools.edit.support.EditGeomPointIterator;
import net.refractions.udig.tools.edit.support.Point;
import net.refractions.udig.tools.edit.support.PrimitiveShape;
import net.refractions.udig.tools.edit.support.ShapeType;
import net.refractions.udig.ui.graphics.ViewportGraphics;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jface.preference.IPreferenceStore;

/**
 * Draws each vertex point as a rectangle.  Listens to mouse events as long as valid and fills vertext when
 * cursor is over a vertex.
 * 
 * @author jones
 * @since 1.1.0
 */
public class DrawPointCommand extends AbstractDrawCommand implements MapMouseMotionListener {
    
    private PrimitiveShape shape;
    private int radius = 3;
    private Color outline = Color.BLACK;
    private Color fill = new Color( 255,168,168 );
    private ViewportPane pane;
    private Point location;
    private int radiusDelta=1;
    private Color selectionFill=Color.YELLOW;
    private boolean overPoint;
    private EditToolHandler handler;
    private boolean drawCurrentShape; 
    IPreferenceStore store = EditPlugin.getDefault().getPreferenceStore();


    public DrawPointCommand( EditToolHandler handler2, PrimitiveShape shape, ViewportPane pane ){
        this.shape=shape;
        this.pane=pane;
        pane.addMouseMotionListener(this);
        this.handler=handler2;
    }

    public void run( IProgressMonitor monitor ) throws Exception {
        graphics.setStroke(ViewportGraphics.LINE_SOLID, 1);
        if( drawCurrentShape ){
            shape=handler.getCurrentShape();
        }
        if( shape==null || shape.getNumPoints()==0)
            return;
        Rectangle rect=new Rectangle();
        rect.width=radius*2;
        rect.height=radius*2;
        boolean selected;
        for( Iterator<Point> iter=new EditGeomPointIterator(shape.getEditGeom()); iter.hasNext();) {
            Point p=iter.next();
            if( (p.getX()<0||p.getX()>display.getWidth()) 
                    && (p.getY()<0 || p.getY()>display.getHeight()) )
                continue;
            rect.x=p.getX()-radius;
            rect.y=p.getY()-radius;
            
            setZoomedRect(rect, p, 0);                    
            if( shape.getEditBlackboard().getSelection().contains(p) ){
                graphics.setColor(selectionFill);
                selected=true;
            }else{
                graphics.setColor(fill);  
                selected=false;
            }
            if( selected || store.getBoolean(PreferenceConstants.P_FILL_VERTICES) 
                    || shape.getEditGeom().getShapeType()==ShapeType.POINT )
                graphics.fillRect(rect.x, rect.y, rect.width, rect.height);
            
            graphics.setColor(outline);
            graphics.drawRect(rect.x, rect.y, rect.width, rect.height);
        }            
        if( overPoint ){
            Point firstPoint = shape.getPoint(0);
            rect.x=firstPoint.getX()-radius;
            rect.y=firstPoint.getY()-radius;
            
            setZoomedRect(rect, firstPoint, radiusDelta);
            if( shape.getEditBlackboard().getSelection().contains(firstPoint) )
                graphics.setColor(selectionFill);
            else
                graphics.setColor(fill);
            graphics.fillRect(rect.x, rect.y, rect.width, rect.height);
            
            graphics.setColor(outline);
            graphics.drawRect(rect.x, rect.y, rect.width, rect.height);
        }

    }

    public void setDrawCurrentShape(boolean b) {
        drawCurrentShape=b;
    }
    
    
    /**
     * Sets the size of the point box.  
     *
     * @param rect rectangle to set
     * @param p center point
     * @param widthDelta the number of pixels larger the box should be than the radius.  
     * (radius of box is size of box from center to edge) delta is number of pixel from edge of normal
     * rect to new edge.  So 1 pixel delta would make rectangle 2 pixels larger. 
     */
    private void setZoomedRect( Rectangle rect, Point p, int widthRad ) {
        rect.width=(widthRad*2)+radius*2;
        rect.height=rect.width;
        int i=radius+widthRad;
        rect.x=p.getX()-i;
        rect.y=p.getY()-i;
    }

    private boolean isOverPoint( Point p ) {
        for( int x=-radiusDelta-radius, max=radius+radiusDelta; x<max; x++){
            for( int y=-radiusDelta-radius; y<max; y++){
                if ( Point.valueOf(p.getX()+x,p.getY()+y).equals(location))
                    return true;
            }            
        }
        return false;
    }
    
    @Override
    public void setValid( boolean valid ) {
        if( !valid ){
            pane.removeMouseMotionListener(this);
        }
        super.setValid(valid);
    }

    public void mouseMoved( MapMouseEvent event ) {
        boolean oldIsOver=overPoint;
        this.location=Point.valueOf(event.x, event.y);

        if( shape==null || shape.getNumPoints()==0 )
            return;
        
        Point firstPoint = shape.getPoint(0);
        overPoint=isOverPoint(firstPoint);
        if( overPoint!=oldIsOver )
            handler.repaint();
    }

    public void mouseDragged( MapMouseEvent event ) {
        mouseMoved(event);
    }

    public void setCurrentShape( PrimitiveShape shape) {
        if( shape==this.shape)
            return;
        this.shape=shape;
        handler.repaint();
    }
    /**
     * 
     * @return Returns the radiusDelta.
     */
    public int getRadiusDelta() {
        return radiusDelta;
    }
    /**
     * Default is 1.
     * @param width_rad The radiusDelta to set.
     */
    public void setRadiusDelta( int radiusDelta ) {
        this.radiusDelta = radiusDelta;
    }
    /**
     * @return Returns the fill.
     */
    public Color getFill() {
        return fill;
    }
    /**
     * Default is Color( 255,168,168,188 )
     * @param fill The fill to set.
     */
    public void setFill( Color fill ) {
        this.fill = fill;
        if( fill.getAlpha()!=255 ){
            this.fill=new Color(fill.getRGB());
        }
    }
    /**
     * @return Returns the outline.
     */
    public Color getOutline() {
        return outline;
    }
    /**
     * Default is Black
     * @param outline The outline to set.
     */
    public void setOutline( Color outline ) {
        this.outline = outline;
    }
    /**
     * @return Returns the radius.
     */
    public int getRadius() {
        return radius;
    }
    /**
     * Default is 3.
     * @param radius The radius to set.
     */
    public void setRadius( int radius ) {
        this.radius = radius-radiusDelta;
    }
    /**
     * @return Returns the geom.
     */
    public PrimitiveShape getGeom() {
        return shape;
    }

    /**
     * @return Returns the selectionFill.
     */
    public Color getSelectionFill() {
        return selectionFill;
    }

    /**
     * @param selectionFill The selectionFill to set.
     */
    public void setSelectionFill( Color selectionFill ) {
        this.selectionFill = selectionFill;
    }

    public Rectangle getValidArea() {
        if( shape!=null )
            return shape.getBounds();
        else
            return null;
    }

    public void mouseHovered( MapMouseEvent event ) {
    }

}
