/* uDig - User Friendly Desktop Internet GIS client
 * http://udig.refractions.net
 * (C) 2004, Refractions Research Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation;
 * version 2.1 of the License.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 */
package net.refractions.udig.tools.edit.commands;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Stack;

import net.refractions.udig.project.ILayer;
import net.refractions.udig.project.command.AbstractCommand;
import net.refractions.udig.project.command.UndoableMapCommand;
import net.refractions.udig.project.command.factory.EditCommandFactory;
import net.refractions.udig.project.ui.AnimationUpdater;
import net.refractions.udig.project.ui.IAnimation;
import net.refractions.udig.tool.edit.internal.Messages;
import net.refractions.udig.tools.edit.EditPlugin;
import net.refractions.udig.tools.edit.EditState;
import net.refractions.udig.tools.edit.EditToolHandler;
import net.refractions.udig.tools.edit.animation.GeometryOperationAnimation;
import net.refractions.udig.tools.edit.behaviour.WriteChangesBehaviour;
import net.refractions.udig.tools.edit.support.EditBlackboard;
import net.refractions.udig.tools.edit.support.EditGeom;
import net.refractions.udig.tools.edit.support.EditUtils;
import net.refractions.udig.tools.edit.support.GeometryCreationUtil;
import net.refractions.udig.tools.edit.support.IsBusyStateProvider;
import net.refractions.udig.tools.edit.support.PrimitiveShape;
import net.refractions.udig.tools.edit.support.ShapeType;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.geotools.data.FeatureSource;
import org.geotools.feature.Feature;
import org.geotools.feature.FeatureCollection;
import org.geotools.feature.FeatureIterator;
import org.geotools.filter.FilterFactoryFinder;

import com.vividsolutions.jts.geom.Coordinate;
import com.vividsolutions.jts.geom.Geometry;
import com.vividsolutions.jts.geom.Polygon;

/**
 * This command splits all the geometries in the Editblackboard based on the Current Shape ([@link
 * net.refractions.udig.tools.edit.EditToolHandler#getCurrentShape()})
 * 
 * @author jones
 * @since 1.1.0
 */
public class SplitFeatureCommand extends AbstractCommand implements UndoableMapCommand {
    private static final String NEW_ID = "newSplitFeature"; //$NON-NLS-1$
    private EditToolHandler handler;
    private PrimitiveShape shape;
    private EditState state;
    private ILayer layer;
    private List<EditGeom> geoms;
    private EditState endState;
    private WriteChangesBehaviour behaviour;
    private UndoableMapCommand writeCommand;
    private List<UndoableMapCommand> createCommands = new ArrayList<UndoableMapCommand>();

    /**
     * @param handler
     */
    public SplitFeatureCommand( EditToolHandler handler, EditState endState ) {
        this.handler = handler;
        this.layer = handler.getEditLayer();
        this.endState = endState;
        behaviour = new WriteChangesBehaviour(Polygon.class);
        behaviour.setUpdateBlackboard(false);
    }
    @SuppressWarnings("unchecked")
    public void run( IProgressMonitor monitor ) throws Exception {
        try {
            handler.setCurrentState(EditState.BUSY);
            shape = handler.getCurrentShape();
            EditBlackboard editBlackboard = handler.getEditBlackboard(layer);
            geoms = editBlackboard.getGeoms();

            if( EditPlugin.isDebugging(EditPlugin.RUN_ASSERTIONS))
                for( EditGeom g : geoms ) {
                    g.assertValid();
                }
            monitor.beginTask(Messages.SplitFeatureCommand_name, geoms.size() * 11 + 1);
            monitor.worked(1);

            editBlackboard.clear();
            
            state = handler.getCurrentState();
            for( EditGeom geom : geoms ) {
                if (geom != handler.getCurrentGeom()) {
                    IAnimation anim = new GeometryOperationAnimation(geom.getShell(),new IsBusyStateProvider(handler));
                    try {
                        AnimationUpdater.runTimer(handler.getContext().getMapDisplay(), anim);
                        List<EditGeom> newGeoms = splitGeom(Collections.singletonList(geom), shape,
                                0);
                        FeatureSource source = layer.getResource(FeatureSource.class,
                                new NullProgressMonitor());
                        FeatureCollection features = source.getFeatures(FilterFactoryFinder
                                .createFilterFactory()
                                .createFidFilter(geom.getFeatureIDRef().get()));
                        FeatureIterator iter = features.features();
                        Feature feature;
                        try {
                            feature = iter.next();
                        } catch (NoSuchElementException e) {
                            feature = null;
                        } finally {
                            iter.close();
                        }
                        if (feature != null) {
                            for( EditGeom geom2 : newGeoms ) {
                                if (feature != null
                                        && geom2.getFeatureIDRef().get().equals(
                                                geom.getFeatureIDRef().get()))
                                    continue;

                                Feature newFeature = feature.getFeatureType().duplicate(feature);
                                Class<Geometry> type = GeometryCreationUtil.determineGeometryType(
                                        handler.getCurrentGeom(), geom2, Polygon.class, feature
                                                .getFeatureType().getDefaultGeometry());

                                Geometry createGeom = GeometryCreationUtil.createGeom(type, geom2
                                        .getShell());
                                newFeature.setDefaultGeometry(GeometryCreationUtil
                                        .ceateGeometryCollection(Collections
                                                .singletonList(createGeom), feature
                                                .getFeatureType().getDefaultGeometry().getType()));
                                UndoableMapCommand command = EditCommandFactory.getInstance()
                                        .createAddFeatureCommand(newFeature, layer);
                                this.createCommands.add(command);
                                command.setMap(getMap());
                                command.run(new NullProgressMonitor());
                                geom2.setChanged(false);
                            }
                        }
                    } finally {
                        anim.setValid(false);
                    }
                }
                monitor.worked(1);

            }

            writeCommand = behaviour.getCommand(handler);
            writeCommand.setMap(getMap());
            writeCommand.run(new SubProgressMonitor(monitor, 10 * geoms.size()));

            editBlackboard.clear();
        } finally {
            handler.setCurrentShape(null);
            handler.setCurrentState(endState);
            monitor.done();
        }
    }

    /**
     * Splits a Polygon against a line. The created geometries are added to the edit blackboard.
     * 
     * @param toSplit geom to split
     * @param splitter shape to use to determine the split.
     * @return
     */
    public static List<EditGeom> splitGeom( List<EditGeom> toSplit, final PrimitiveShape splitter,
            final int edgeIndex ) {
        if (splitter.getNumCoords() == 1) {
            throw new IllegalArgumentException("Splitting shape must have > 1 coordinates"); //$NON-NLS-1$
        }
        List<EditGeom> result = new ArrayList<EditGeom>();

        for( EditGeom geom : toSplit ) {
            if (geom.getShell().getNumCoords() == 1) {
                continue;
            }
            List<EdgeIntersection> intersections = new ArrayList<EdgeIntersection>();
            int i = edgeIndex;
            for( ; i < splitter.getNumCoords() - 1
                    && (intersections.size() < 2 || intersections.get(0).split.shape != intersections
                            .get(1).split.shape); i++ ) {
                final Edge splitterEdge = new Edge(i, i + 1, splitter);

                findIntersections(geom, splitterEdge, intersections);

            }
            Collections.sort(intersections, new Comparator<EdgeIntersection>(){
                public int compare( EdgeIntersection o1, EdgeIntersection o2 ) {
                    EditUtils.MinFinder finder = new EditUtils.MinFinder(splitter
                            .getCoord(edgeIndex));
                    finder.add(o1.intersectionCoord);
                    finder.add(o2.intersectionCoord);
                    return finder.getMinCoord().equals(o1.intersectionCoord) ? -1 : 1;
                }
            });

            intersections = removeIntersectionsRunningOneEdge(geom, intersections);

            if (intersections.size() < 2) {
                result.add(geom);
                continue;
            }

            result.addAll(split(geom, intersections));

            if (i < splitter.getNumCoords() - 2)
                result = splitGeom(result, splitter, i);
        }

        return result;
    }

    private static List<EdgeIntersection> removeIntersectionsRunningOneEdge( EditGeom geom,
            List<EdgeIntersection> intersections ) {
        if (intersections.size() < 2)
            return intersections;
        List<EdgeIntersection> goodIntersections = new ArrayList<EdgeIntersection>();
        Iterator<EdgeIntersection> iter = intersections.iterator();
        EdgeIntersection intersection = iter.next();
        goodIntersections.add(intersection);
        while( iter.hasNext() ) {
            EdgeIntersection next = iter.next();
            if (intersection.split.shape != next.split.shape) {
                goodIntersections.add(next);
                intersection = next;
                continue;
            }

            if (next.intersectionCoord.equals(intersection.intersectionCoord)) {
                continue;
            }
            if (intersection.split.equals(next.split)
                    && intersection.splitter.equals(next.splitter)) {
                continue;
            }

            int start = 0;
            int end = intersection.split.shape.getNumCoords() - 1;

            Edge edge;
            boolean doBreak = false;
            while( start < end ) {
                edge = new Edge(start, start + 1, intersection.split.shape);
                if (edge.contains(intersection.intersectionCoord) && edge.contains(next.intersectionCoord)
                        && intersection.splitter.equals(next.splitter)) {
                    goodIntersections.remove(intersection);
                    goodIntersections.add(next);
                    intersection = next;
                    doBreak = true;
                    break;
                }
                start++;
            }

            if (!doBreak) {
                goodIntersections.add(next);
                intersection = next;
            }
        }

        return goodIntersections;
    }

    private static List<EditGeom> split( EditGeom toSplit, List<EdgeIntersection> intersections ) {

        if (intersections.size() < 2)
            return Collections.singletonList(toSplit);
        
        List<EdgeIntersection> results = intersections;
        
        EditBlackboard bb = toSplit.getEditBlackboard();
        bb.removeGeometries(Collections.singleton(toSplit));
        ShapeType shapeType = toSplit.getShapeType();
        EditGeom left = bb.newGeom(toSplit.getFeatureIDRef().get(), shapeType);
        left.setChanged(true);

        // mark this feature as a new feature
        // time stamp is to ensure that there isn't another feature in feature source with same
        // id.
        EditGeom right = bb.newGeom(NEW_ID + System.currentTimeMillis(), shapeType);

        ArrayList<EditGeom> list = new ArrayList<EditGeom>();

        final int OUTSIDE = 0;
        final int IN_SHAPE = 1;
        final int IN_HOLE = 2;
        int state = OUTSIDE;

        // this is to keep track of the last shape entered.
        Stack<EdgeIntersection> oldIntersections = new Stack<EdgeIntersection>();
        Edge startShape = null;
        EdgeIntersection startHole = null;
        int lastState = -1;

        for( Iterator<EdgeIntersection> iter = results.iterator(); iter.hasNext(); ) {
            EdgeIntersection intersection = iter.next();
            int tmp = state;
            switch( state ) {
            case OUTSIDE: {
                state = IN_SHAPE;
                startShape = intersection.split;
                if (lastState == IN_SHAPE) {
                    EditGeom geom;

                    if (left.getShell().hasVertex(intersection.split.start)
                            && left.getShell().hasVertex(intersection.split.end)) {
                        geom = left;
                        addToResults(right, list);
                    } else if (right.getShell().hasVertex(intersection.split.start)
                            && right.getShell().hasVertex(intersection.split.end)) {
                        geom = right;
                        addToResults(left, list);
                    } else {
                        addToResults(right, list);
                        addToResults(left, list);
                        return list;
                    }
                    for( EdgeIntersection intersection2 : results ) {
                        intersection2.split.shape = geom.getShell();
                        Edge edge = intersection2.split;
                        try{
                        edge.startIndex = match(toSplit.getShell(), geom.getShell(), edge);
                        }catch (RuntimeException e) {
                            addToResults(geom, list);
                            return list;
                        }
                        edge.endIndex = edge.startIndex + 1;
                    }
                    list.addAll(split(geom, results));
                    return list;
                }
                addCoord(intersection.intersectionCoord, left.getShell());
                addCoord(intersection.intersectionCoord, right.getShell());
                oldIntersections.push(intersection);
                break;
            }
            case IN_SHAPE: {
                if (intersection.split.shape == oldIntersections.peek().split.shape) {
                    // exiting shape.
                    state = OUTSIDE;
                    if (lastState != IN_HOLE)
                        for( int i = oldIntersections.peek().splitter.endIndex; i < intersection.splitter.endIndex; i++ ) {
                            Coordinate coord = intersection.splitter.shape.getCoord(i);
                            bb.addCoordinate(coord, left.getShell());
                            bb.addCoordinate(coord, right.getShell());
                        }
                    addCoord(intersection.intersectionCoord, left.getShell());
                    addCoord(intersection.intersectionCoord, right.getShell());
                    // add coordinates to right
                    if (!intersection.split.equals(oldIntersections.peek().split)) {
                        if (intersection.intersectionCoord.equals(intersection.split.start)) {
                            addRestOfCoords(false, intersection.split.startIndex - 1,
                                    startShape.endIndex, right.getShell(),
                                    oldIntersections.peek().split.shape);
                        } else {
                            addRestOfCoords(false, intersection.split.startIndex,
                                    startShape.endIndex, right.getShell(),
                                    oldIntersections.peek().split.shape);
                        }
                    }
                    // add coordinates to left
                    if (intersection.intersectionCoord.equals(intersection.split.end)) {
                        addRestOfCoords(true, intersection.split.endIndex + 1,
                                startShape.startIndex, left.getShell(),
                                oldIntersections.peek().split.shape);
                    } else {
                        addRestOfCoords(true, intersection.split.endIndex, startShape.startIndex,
                                left.getShell(), oldIntersections.peek().split.shape);
                    }

                    bb.addCoordinate(oldIntersections.peek().intersectionCoord, left.getShell());
                    bb.addCoordinate(oldIntersections.peek().intersectionCoord, right.getShell());
//                    System.out.println("original: "+Arrays.asList(toSplit.getShell().coordArray()));
//                    System.out.println("left: "+Arrays.asList(left.getShell().coordArray()));
//                    System.out.println("right: "+Arrays.asList(right.getShell().coordArray()));
                    EditUtils.instance.reverseOrder(right.getShell());
//                    System.out.println("After reorder: "+Arrays.asList(right.getShell().coordArray()));
                    oldIntersections.pop();
                    left.getShell().assertValid();
                    right.getShell().assertValid();
                } else {
                    // entering hole
                    state = IN_HOLE;

                    startHole = intersection;

                    addCoord(intersection.intersectionCoord, left.getShell());
                    addCoord(intersection.intersectionCoord, right.getShell());
                    oldIntersections.push(intersection);
                }
                break;
            }
            case IN_HOLE: {
                state = IN_SHAPE;
                // add coordinates to right
                addRestOfCoords(false, startHole.split.startIndex, intersection.split.endIndex,
                        right.getShell(), oldIntersections.peek().split.shape);
                // add coordinates to left
                addRestOfCoords(true, startHole.split.endIndex, intersection.split.startIndex, left
                        .getShell(), oldIntersections.peek().split.shape);

                addCoord(intersection.intersectionCoord, left.getShell());
                addCoord(intersection.intersectionCoord, right.getShell());
                oldIntersections.pop();
                break;
            }

            default:
                throw new IllegalStateException("Only legal states are 0,1,2"); //$NON-NLS-1$
            }
            lastState = tmp;
            iter.remove();
        }

        for( PrimitiveShape hole : toSplit.getHoles() ) {
            boolean found = false;
            for( EdgeIntersection intersection : results ) {
                if (hole == intersection.split.shape) {
                    found = true;
                }
            }
            if (!found) {
                if (hole.getNumCoords() > 0) {
                    Coordinate p = hole.getCoord(0);
                    if (left.getShell().contains(p.x, p.y)) {
                        PrimitiveShape newHole = left.newHole();
                        for( Iterator<Coordinate> iter=hole.coordIterator(); iter.hasNext(); ) {
                            Coordinate coord = iter.next();
                            bb.addCoordinate(coord, newHole);
                        }
                    } else {
                        PrimitiveShape newHole = right.newHole();
                        for( Iterator<Coordinate> iter=hole.coordIterator(); iter.hasNext(); ) {
                            Coordinate coord = iter.next();
                            bb.addCoordinate(coord, newHole);
                        }
                    }
                }
            }
        }

        addToResults(right, list);
        addToResults(left, list);

        return list;
    }
    /**
     *
     * @param right
     * @param list
     */
    private static void addToResults( EditGeom right, ArrayList<EditGeom> list ) {
        if( right.getShell().getNumCoords()<4)
            return;
        list.add(right);
    }

    private static int match( PrimitiveShape from, PrimitiveShape to, Edge edge ) {
        int index = matchForward(from, to, edge);
        if (index == -1)
            index = matchBackward(from, to, edge);
        return index;
    }
    private static int matchBackward( PrimitiveShape from, PrimitiveShape to, Edge edge ) {
        for( int i = edge.startIndex - 1; i > 0; i-- ) {
            if (isMatch(from, to, edge, i))
                return i;
        }
        throw new RuntimeException("Edge: " + edge.start + ":" + edge.end + " not found in " + to); //$NON-NLS-1$//$NON-NLS-2$//$NON-NLS-3$
    }
    private static int matchForward( PrimitiveShape from, PrimitiveShape to, Edge edge ) {
        for( int i = edge.startIndex; i < to.getNumCoords() - 1; i++ ) {
            if (isMatch(from, to, edge, i))
                return i;
        }
        return -1;
    }
    private static boolean isMatch( PrimitiveShape from, PrimitiveShape to, Edge edge, int i ) {
        if (from.getCoord(edge.startIndex).equals(to.getCoord(i))) {
            if (i + 1 < to.getNumCoords() - 1 && i + 1 < from.getNumCoords() - 1
                    && from.getCoord(edge.startIndex + 1).equals(to.getCoord(i + 1)))
                return true;
            else if (i - 1 > 0 && from.getCoord(edge.startIndex - 1).equals(to.getCoord(i - 1)))
                return true;
            return false;
        } else {
            if (i + 2 < to.getNumCoords() - 1 && i + 2 < from.getNumCoords() - 1
                    && from.getCoord(edge.startIndex + 1).equals(to.getCoord(i + 1))
                    && from.getCoord(edge.startIndex + 2).equals(to.getCoord(i + 2)))
                return true;
            return false;
        }
    }
    private static void addCoord( Coordinate coord, PrimitiveShape shape ) {
        EditBlackboard bb = shape.getEditBlackboard();
        int coordinates = shape.getNumCoords();
        if (coordinates > 1) {
            Coordinate backTwo = shape.getCoord(coordinates - 2);
            Coordinate backOne = shape.getCoord(coordinates - 1);
            if (((Math.abs(backTwo.x - coord.x) < 0.00000001 && Math.abs(backOne.x - coord.x) < 0.00000001))
                    || ((Math.abs(backTwo.y - coord.y) < 0.00000001 && Math
                            .abs(backOne.y - coord.y) < 0.00000001))) {
                
                bb.removeCoordinate(shape.getAssociatedPointIndex(coordinates - 1), backOne, shape);
            }
        }
        bb.addCoordinate(coord, shape);
    }
    private static void findIntersections( EditGeom toSplit, final Edge splitterEdge,
            List<EdgeIntersection> intersections ) {
        for( PrimitiveShape shape : toSplit ) {
            for( int j = 0; j < shape.getNumCoords() - 1; j++ ) {
                final Coordinate start = shape.getCoord(j);
                Coordinate end = shape.getCoord(j + 1);

                Coordinate intersection = EditUtils.instance.intersectingLines(splitterEdge.start,
                        splitterEdge.end, start, end);
                if (intersection != null) {
                    EdgeIntersection ei = new EdgeIntersection();
                    ei.intersectionCoord = intersection;
                    ei.split = new Edge(j, j + 1, shape);
                    ei.splitter = splitterEdge;
                    intersections.add(ei);
                }
            }
        }

    }

    private static void addRestOfCoords( boolean increasingOrder, int start, int end,
            PrimitiveShape toShape, PrimitiveShape fromShape ) {
        int increment = 1;
        int stop = end;
        if (!increasingOrder) {
            increment = -1;
            if (end > start)
                stop = 0;
        } else {
            if (end < start)
                stop = fromShape.getNumCoords() - 1;
        }
        for( int i = start; i != stop + increment
                && !toShape.getCoord(0).equals(fromShape.getCoord(i)); i += increment ) {
                addCoord(fromShape.getCoord(i), toShape);
        }

        if (stop == end)
            return;
        int newStart = 0;

        if (!increasingOrder) {
            newStart = fromShape.getNumCoords() - 1;
        }

        for( int i = newStart; i != end + increment
                && !toShape.getCoord(0).equals(fromShape.getCoord(i)); i += increment ) {
                    addCoord(fromShape.getCoord(i), toShape);
        }
    }

    static class EdgeIntersection {
        Coordinate intersectionCoord;
        Edge splitter;
        Edge split;

        @Override
        public String toString() {
            return intersectionCoord.toString();
        }
    }

    static class Edge {
        PrimitiveShape shape;
        Coordinate start;
        Coordinate end;
        int startIndex;
        int endIndex;

        public Edge( int startIndex2, int endIndex2, PrimitiveShape shape2 ) {
            this.startIndex = startIndex2;
            this.endIndex = endIndex2;
            this.start = shape2.getCoord(startIndex2);
            this.end = shape2.getCoord(endIndex2);
            this.shape = shape2;
        }

        public boolean contains( Coordinate coordinate ) {
            Coordinate closestPointOnEdge = EditUtils.instance.closestCoordinateOnEdge(start, end, coordinate);
            if (closestPointOnEdge != null && closestPointOnEdge.equals(coordinate) )
                return true;
            return false;
        }

        @Override
        public int hashCode() {
            int seed = 37;
            seed += 17 * start.hashCode();
            seed += 17 * end.hashCode();
            return seed;
        }
        @Override
        public String toString() {
            return start + "-" + end; //$NON-NLS-1$
        }

        public void init( int startIndex2, int endIndex2 ) {
            this.startIndex = startIndex2;
            this.endIndex = endIndex2;
            this.start = shape.getCoord(startIndex2);
            this.end = shape.getCoord(endIndex2);
        }

        @Override
        public boolean equals( Object obj ) {
            if (!(obj instanceof Edge))
                return false;
            Edge other = (Edge) obj;
            if (other.shape != shape)
                return false;

            return (other.start.equals(start) && other.end.equals(end))
                    || (other.start.equals(end) && other.end.equals(start));
        }
    }

    public void rollback( IProgressMonitor monitor ) throws Exception {
        try {
            int ticks = (this.createCommands.size() + 1) * 10 + geoms.size() + 1;
            monitor.beginTask(Messages.SplitFeatureCommand_undoMessage, ticks);
            monitor.worked(1);

            writeCommand.rollback(new SubProgressMonitor(monitor, 10));
            for( UndoableMapCommand command : this.createCommands ) {
                command.rollback(new SubProgressMonitor(monitor, 10));
            }
            handler.setCurrentState(EditState.BUSY);
            EditBlackboard bb = handler.getEditBlackboard(layer);
            bb.clear();
            for( EditGeom geom : geoms ) {
                EditGeom newGeom = bb.newGeom(geom.getFeatureIDRef().get(), geom.getShapeType());
                newGeom.setChanged(geom.isChanged());
                for( PrimitiveShape shape : geom ) {
                    Coordinate[] coords = shape.coordArray();
                    for( int i = 0; i < coords.length; i++ ) {
                        bb.addCoordinate(coords[i], shape);
                    }
                }
                monitor.worked(1);
            }
        } finally {
            handler.setCurrentState(state);
            monitor.done();
        }
    }

    public String getName() {
        return Messages.SplitFeatureCommand_name;
    }

}
